/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
#include "cmogstored.h"
#include "compat_accept.h"

MOG_NOINLINE static void accept_error_check(struct mog_accept *ac)
{
	int fd;

	switch (errno) {
	case ECONNABORTED:
	case EINTR:
		return; /* common errors, nothing we can do about it */
	case EBADF:
		/* can happen due to unavoidable race at shutdown */
	case ENOTSOCK:
	case EOPNOTSUPP:
		return pthread_exit(NULL);
	case_EAGAIN:
		fd = mog_fd_of(ac)->fd;
		if (set_nonblocking_flag(fd, false) != 0) {
			if (errno == EBADF)
				return pthread_exit(NULL);
			syslog(LOG_ERR,
			       "failed to make fd=%d blocking: %m", fd);
		}
		syslog(LOG_DEBUG, "made fd=%d blocking", fd);
		return;
	case EMFILE:
	case ENFILE:
	case ENOBUFS:
	case ENOMEM:
		syslog(LOG_ERR, "accept4() failed with: %m");
		return;
	default:
		syslog(LOG_ERR, "accept4() failed with: %m");
	}
}

static void post_accept_run(struct mog_accept *ac, int client_fd)
{
	struct mog_fd *mfd = mog_fd_get(client_fd);

	mfd->fd = client_fd;
	mfd->queue_state = MOG_QUEUE_STATE_NEW;
	ac->post_accept_fn(mfd, ac->svc);
}

/*
 * passed as the start_routine argument to pthread_create.
 * This function may run concurrently in multiple threads.
 * The design of cmogstored assumes "wake-one" behavior for blocking
 * accept()/accept4() callers.  We will force accept_fd into blocking
 * state if O_NONBLOCK is ever set (e.g. listen socket was inherited).
 */
void *mog_accept_loop(void *arg)
{
	struct mog_accept *ac = arg;
	int accept_fd = mog_fd_of(ac)->fd;

	for (;;) {
		/* pthread cancellation point */
		int client_fd = mog_accept(accept_fd, NULL, NULL);

		if (client_fd >= 0) {
			mog_cancel_disable();
			post_accept_run(ac, client_fd);
			mog_cancel_enable();
		} else {
			accept_error_check(ac);
		}
	}

	return NULL;
}
