/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
/*
 * common headers, macros, and static inline functions for the entire project
 *
 * Internal APIs are very much in flux and subject to change frequently
 */
#include "config.h"
#include "queue_kqueue.h"
#define _XOPEN_SOURCE 700
#define  _POSIX_C_SOURCE 200809L
#include <pthread.h>
#include <sys/uio.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <signal.h>
#include <stddef.h>
#include <stdint.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <errno.h>
#include <limits.h>
#include <dirent.h>
#include <syslog.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/statvfs.h>
#include <time.h>
#include <argp.h>
#include "bsd/queue_safe.h"
#include "bsd/simpleq.h"

/* gnulib headers */
#include "hash.h"
#include "xalloc.h"
#include "xvasprintf.h"
#include "nonblocking.h"
#include "canonicalize.h"
#include "verify.h"
#include "cloexec.h"
#include "mountlist.h"
#include "base64.h"
#include "minmax.h"
#include "md5.h"

#include "gcc.h"
#include "util.h"
#include "defaults.h"
#include "iostat.h"
#include "mnt.h"

#define MOG_WR_ERROR ((void *)-1)
#define MOG_IOSTAT (MAP_FAILED)
#define MOG_FD_MAX (INT_MAX-1)

enum mog_write_state {
	MOG_WRSTATE_ERR = -1,
	MOG_WRSTATE_DONE = 0,
	MOG_WRSTATE_BUSY = 1
};

enum mog_parser_state {
	MOG_PARSER_ERROR = -1,
	MOG_PARSER_DONE = 0,
	MOG_PARSER_CONTINUE = 1
};

enum mog_next {
	MOG_NEXT_CLOSE = 0,
	MOG_NEXT_ACTIVE,
	MOG_NEXT_WAIT_RD,
	MOG_NEXT_WAIT_WR,
	MOG_NEXT_IGNORE /* for iostat */
};

struct mog_wbuf;
struct mog_dev {
	dev_t st_dev;
	uint32_t devid;
	char prefix[FLEXIBLE_ARRAY_MEMBER];
};

struct mog_rbuf {
	size_t rsize;
	char rptr[FLEXIBLE_ARRAY_MEMBER];
};

/*
 * overhead based on the malloc implementation found in glibc,
 * it's unlikely another memalign implentation would have _more_ overhead...
 */
#define MOG_MEMALIGN_OVERHEAD (sizeof(size_t) * 2)
#define MOG_RBUF_OVERHEAD (sizeof(struct mog_rbuf) + MOG_MEMALIGN_OVERHEAD)

struct mog_mgmt;
struct mog_mgmt {
	int cs;
	struct mog_fd *forward;
	size_t offset;
	size_t mark[2];
	struct mog_rbuf *rbuf;
	struct mog_wbuf *wbuf; /* uncommonly needed */
	struct mog_svc *svc;
	LIST_ENTRY(mog_mgmt) subscribed;
};

struct mog_queue;
struct mog_svc {
	int dirfd;
	const char *docroot;

	/* private */
	DIR *dir;
	Hash_table *by_st_dev;
	pthread_mutex_t devstats_lock;
	struct mog_queue *queue;
	LIST_HEAD(mgmt_head, mog_mgmt) devstats_subscribers;
	mode_t put_perms;
	mode_t mkcol_perms;
	int http_fd;
	int httpget_fd;
	int mgmt_fd;
};

enum mog_http_method {
	MOG_HTTP_METHOD_NONE = 0,
	MOG_HTTP_METHOD_GET,
	MOG_HTTP_METHOD_HEAD,
	MOG_HTTP_METHOD_PUT,
	MOG_HTTP_METHOD_DELETE,
	MOG_HTTP_METHOD_MKCOL
};

enum mog_chunk_state {
	MOG_CHUNK_STATE_SIZE = 0,
	MOG_CHUNK_STATE_DATA,
	MOG_CHUNK_STATE_TRAILER,
	MOG_CHUNK_STATE_DONE
};

struct mog_http {
	int cs;
	enum mog_http_method http_method:8;
	unsigned get_only:1;
	unsigned persistent:1;
	unsigned chunked:1;
	unsigned has_trailer_md5:1;
	unsigned has_expect_md5:1;
	unsigned has_content_range:1; /* for PUT */
	unsigned has_range:1;         /* for GET */
	unsigned skip_rbuf_defer:1;
	enum mog_chunk_state chunk_state:2;
	uint8_t path_tip;
	uint8_t path_end;
	uint16_t line_end;
	uint16_t tmp_tip;
	struct mog_fd *forward;
	size_t offset;
	off_t range_beg;
	off_t range_end;
	off_t content_len;
	struct mog_rbuf *rbuf;
	struct mog_wbuf *wbuf; /* uncommonly needed */
	struct mog_svc *svc;
	uint8_t expect_md5[16];
};

struct mog_thrpool {
	size_t n_threads;
	pthread_t *threads;
};

struct mog_fd;

/*
 * this is a two-part queue: epoll or kqueue for the idle blocking part,
 * and SIMPLEQ is the non-blocking part.
 * mog_queue objects can be shared by any number of mog_svcs
 */
struct mog_queue {
	int queue_fd; /* epoll or kqueue */
	SIMPLEQ_HEAD(active_fd, mog_fd) activeq_head;
	pthread_mutex_t activeq_lock; /* may be pointer for non-NPTL */
	struct mog_thrpool thrpool;
};

/* accept.c */
typedef void (*post_accept_fn)(struct mog_fd *, struct mog_svc *);
struct mog_accept {
	struct mog_svc *svc;
	post_accept_fn post_accept_fn;
	struct mog_thrpool thrpool;
};
struct mog_accept * mog_accept_init(int fd, struct mog_svc *, post_accept_fn);
void * mog_accept_loop(void *ac);

struct md5_ctx;
struct mog_file {
	off_t fsize;
	off_t foff;
	char *path;
	size_t pathlen;
	char *tmppath;
	void *mmptr;
	struct mog_svc *svc;
	struct md5_ctx *md5;
};

enum mog_queue_state {
	MOG_QUEUE_STATE_NEW = 0,
	MOG_QUEUE_STATE_OLD = 1
};

#include "queue_epoll.h"

enum mog_fd_type {
	MOG_FD_TYPE_UNUSED = 0,
	MOG_FD_TYPE_HTTP,
	MOG_FD_TYPE_MGMT,
	MOG_FD_TYPE_ACCEPT,
	MOG_FD_TYPE_FILE,
	MOG_FD_TYPE_QUEUE,
	MOG_FD_TYPE_IOSTAT,
	MOG_FD_TYPE_SVC /* for docroot dirfd */
};

/* fdmap.c */
struct mog_fd {
	enum mog_fd_type fd_type;
	enum mog_queue_state queue_state;
	int fd; /* redundant ... */
	int in_queue;
	SIMPLEQ_ENTRY(mog_fd) active_fd;
	union {
		struct mog_accept accept;
		struct mog_mgmt mgmt;
		struct mog_http http;
		struct mog_file file;
		struct mog_queue queue;
		struct mog_iostat iostat;
		struct mog_svc *svc;
	} as;
};
struct mog_fd *mog_fd_get(int fd);
void mog_fd_put(struct mog_fd *, int fd);
void mog_fdmap_stop_queues(void);
#include "fdmap.h"

/* alloc.c */
void mog_free_and_null(void *ptrptr);
_Noreturn void mog_oom(void);
void *mog_cachealign(size_t size);
struct mog_rbuf *mog_rbuf_new(size_t size);
struct mog_rbuf *mog_rbuf_get(size_t size);
struct mog_rbuf *mog_rbuf_defer(struct mog_rbuf *rbuf);
void mog_rbuf_free(struct mog_rbuf *);
void mog_rbuf_free_and_null(struct mog_rbuf **);
void *mog_fsbuf_get(size_t *size);

/* die.c */
_Noreturn void die(const char *fmt, ...) MOG_PRINTF;

/* warn.c */
void warn(const char *fmt, ...) MOG_PRINTF;

/* maxconns.c */
void mog_set_maxconns(unsigned long);

/* svc.c */
struct mog_svc *mog_svc_new(const char *docroot);
typedef int (*mog_scandev_cb)(const struct mog_dev *, struct mog_svc *);
int mog_svc_scandev(struct mog_svc *, mog_scandev_cb);
size_t mog_svc_each(Hash_processor processor, void *data);

/* dev.c */
struct mog_dev * mog_dev_new(struct mog_svc *, uint32_t mog_devid);
int mog_dev_mkusage(const struct mog_dev *, struct mog_svc *);

/* valid_path.rl */
int mog_valid_path(const char *buf, size_t len);

/* trywrite.c */
void * mog_trywritev(int fd, struct iovec *iov, int iovcnt);
enum mog_write_state mog_tryflush(int fd, struct mog_wbuf **);

/* fs.c */
extern int (*mog_stat)(struct mog_svc *, const char *path, struct stat *sb);
extern int (*mog_unlink)(struct mog_svc *, const char *path);
extern int (*mog_open_read)(struct mog_svc *, const char *path);
extern int (*mog_open_put)(struct mog_svc *, const char *path, int flags);
extern int (*mog_rename)(struct mog_svc *, const char *old, const char *new);
extern int (*mog_mkdir)(struct mog_svc *, const char *path, mode_t mode);
void mog_cloexec_works(void);

/* pidfile.c */
int mog_pidfile_open(const char *path, pid_t *cur);
int mog_pidfile_commit(int fd);

/* svc_dev.c */
bool mog_svc_devstats_broadcast(void *svc, void *ignored);
void mog_svc_devstats_subscribe(struct mog_mgmt *);

/* cloexec_detect.c */
extern bool mog_cloexec_atomic;

/* cloexec_from.c */
void mog_cloexec_from(int lowfd);

/* iostat_process.c */
struct mog_iostat * mog_iostat_spawn(struct mog_queue *);
struct mog_iostat * mog_iostat_get(struct mog_queue *);

/* cfg_parser.rl */
struct mog_cfg;
int mog_cfg_parse(struct mog_cfg *, char *buf, size_t len);

/* cfg.c */
struct mog_cfg * mog_cfg_new(const char *configfile);
int mog_cfg_load(struct mog_cfg *);
void mog_cfg_svc_start_or_die(struct mog_cfg *cli);
extern struct mog_cfg mog_cli;
extern bool mog_cfg_multi;

/* listen_parser.rl */
struct mog_addrinfo *mog_listen_parse(const char *host_with_port);

/* canonpath.c */
char *mog_canonpath(const char *path, enum canonicalize_mode_t canon_mode);
char *mog_canonpath_die(const char *path, enum canonicalize_mode_t canon_mode);

/* thrpool.c */
void mog_thrpool_start(struct mog_thrpool *, size_t n,
                       void *(*start_fn)(void *), void *arg);
void mog_thrpool_quit(struct mog_thrpool *);

/* mgmt.c */
void mog_mgmt_writev(struct mog_mgmt *, struct iovec *, int iovcnt);
void mog_mgmt_post_accept(struct mog_fd *, struct mog_svc *);
void mog_mgmt_queue_step(struct mog_fd *);

/* queue_epoll.c */
struct mog_queue * mog_queue_new(void);
void mog_idleq_push(struct mog_queue *, struct mog_fd *, enum mog_qev);
struct mog_fd * mog_idleq_wait(struct mog_queue *, int dontwait);

/* activeq.c */
void mog_activeq_push(struct mog_queue *, struct mog_fd *mfd);
struct mog_fd *mog_activeq_trytake(struct mog_queue *);

/* addrinfo.c */
struct mog_addrinfo {
	const char *orig;
	struct addrinfo *addr;
};
void mog_addrinfo_free(struct mog_addrinfo **);

/* bind_listen.c */
int mog_bind_listen(struct addrinfo *, const char *accept_filter);

/* close.c */
void mog_close(int fd);

/* mog_queue_loop.c */
void * mog_queue_loop(void *arg);

/* queue_step.c */
void mog_queue_step(struct mog_fd *mfd);

/* sig.c */
void mog_intr_disable(void);
void mog_intr_enable(void);

/* file.c */
struct mog_fd * mog_file_open_read(struct mog_svc *, char *path);
struct mog_fd * mog_file_open_put(struct mog_svc *, char *path, int flags);
void mog_file_close(struct mog_fd *);

/* notify.c */
enum mog_notification {
	MOG_NOTIFY_SIGNAL = -1,
	MOG_NOTIFY_DEVICE_REFRESH = 0,
	MOG_NOTIFY_MAX
};
void mog_notify_init(void);
void mog_notify(enum mog_notification);
void mog_notify_wait(void);

/* http_parser.rl */
void mog_http_reset_parser(struct mog_http *);
void mog_http_init(struct mog_http *, struct mog_svc *);
enum mog_parser_state mog_http_parse(struct mog_http *, char *buf, size_t len);

/* http_get.c */
void mog_http_get_open(struct mog_http *, char *buf, bool head_only);
enum mog_next mog_http_get_in_progress(struct mog_fd *);

/* http.c */
void mog_http_post_accept(struct mog_fd *, struct mog_svc *);
void mog_http_queue_step(struct mog_fd *);
char *mog_http_path(struct mog_http *, char *buf);

/* http_dav.c */
void mog_http_delete(struct mog_http *http, char *buf);
void mog_http_mkcol(struct mog_http *http, char *buf);

/* http_put.c */
void mog_http_put(struct mog_http *http, char *buf, size_t buf_len);
enum mog_next mog_http_put_in_progress(struct mog_fd *);
bool mog_http_write_full(struct mog_fd *file_mfd, char *buf, size_t buf_len);

/* chunk_parser.rl */
void mog_chunk_init(struct mog_http *);
enum mog_parser_state mog_chunk_parse(struct mog_http *, char *buf, size_t len);

/* http_date.c */
#define MOG_HTTPDATE_CAPA (sizeof("Thu, 01 Jan 1970 00:00:00 GMT"))
struct mog_now {
	time_t ntime;
	char httpdate[MOG_HTTPDATE_CAPA];
};
char *mog_http_date(char *dst, size_t len, time_t *timep);
struct mog_now *mog_now(void);
