/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 *
 * All path operations we support are relative to mog_svc.docroot.
 * Absolute path lookups are sometimes more expensive, especially
 * for deeper directory structures, so we'll favor AT_FILE functions.
 *
 * However, it's somewhat common for systems to have AT_FILE functions
 * to be available on the system they're built on, but not available
 * on the system the code is deployed on.  We'll provide fallbacks to
 * the absolute path variants for everything.
 */
#include "cmogstored.h"
#define MY_PATHMAX 80
#ifndef O_NOATIME
# define O_NOATIME 0
#endif

static int noatime_flags = O_RDONLY | O_NOATIME;
static int put_flags = O_RDWR;

/*
 * Like AT_FILE functions, we may be built on systems with O_CLOEXEC and
 * run on systems without it.  O_CLOEXEC is still too recent and was
 * broken on some Linux kernels, so we always run cloexec_detect.c at
 * startup.
 */
#ifdef O_CLOEXEC
void mog_cloexec_works(void)
{
	noatime_flags |= O_CLOEXEC;
	put_flags |= O_CLOEXEC;
}
#endif /* O_CLOEXEC */

/* we only use real *at syscalls, Gnulib workalikes aren't thread-safe */

#define GET_FSPATH(DST,SRC) do { \
	int rc = snprintf((DST), sizeof(DST), "%s%s", svc->docroot, (SRC)); \
	if (rc <= 0 || rc >= sizeof(DST)) { \
		errno = ENAMETOOLONG; \
		return -1; \
	} \
} while (0)

static int mog_statpath(struct mog_svc *svc, const char *path, struct stat *sb)
{
	char fspath[MY_PATHMAX];

	GET_FSPATH(fspath, path);
	return stat(fspath, sb);
}

static int mog_openpath_read(struct mog_svc *svc, const char *path)
{
	int fd;
	char fspath[MY_PATHMAX];

	GET_FSPATH(fspath, path);
	fd = open(fspath, noatime_flags);

	if (fd < 0 && errno != ENOENT && (noatime_flags & O_NOATIME)) {
		noatime_flags = O_RDONLY;
		fd = open(fspath, noatime_flags);
	}
	return fd;
}

static int mog_openpath_put(struct mog_svc *svc, const char *path, int flags)
{
	int fd;
	char fspath[MY_PATHMAX];

	GET_FSPATH(fspath, path);
	fd = open(fspath, flags | put_flags, 0600);

	return fd;
}

#ifdef HAVE_FSTATAT
static int mog_statat(struct mog_svc *svc, const char *path, struct stat *sb)
{
	int rc = fstatat(svc->dirfd, path + 1, sb, 0);

	if (rc < 0 && errno == ENOSYS)
		return mog_statpath(svc, path, sb);
	return rc;
}
#else
#  define mog_statat mog_statpath
#endif

#ifdef HAVE_OPENAT
static int mog_openat_read(struct mog_svc *svc, const char *path)
{
	int rc = openat(svc->dirfd, path + 1, noatime_flags);

	if (rc < 0 && errno == ENOSYS)
		return mog_openpath_read(svc, path);

	if (rc < 0 && errno != ENOENT && (noatime_flags & O_NOATIME)) {
		noatime_flags = O_RDONLY;
		rc = openat(svc->dirfd, path + 1, noatime_flags);
	}

	return rc;
}

static int mog_openat_put(struct mog_svc *svc, const char *path, int flags)
{
	int rc = openat(svc->dirfd, path + 1, flags | put_flags, 0600);

	if (rc < 0 && errno == ENOSYS)
		return mog_openpath_put(svc, path, flags);

	return rc;
}
#else
#  define mog_openat_read mog_openpath_read
#  define mog_openat_put mog_openpath_put
#endif

static int mog_unlinkpath(struct mog_svc *svc, const char *path)
{
	char fspath[MY_PATHMAX];

	GET_FSPATH(fspath,path);
	return unlink(fspath);
}
#ifdef HAVE_UNLINKAT
static int mog_unlinkat(struct mog_svc *svc, const char *path)
{
	int rc = unlinkat(svc->dirfd, path + 1, 0);

	if (rc < 0 && errno == ENOSYS)
		return mog_unlinkpath(svc, path);
	return rc;
}
#else
#  define mog_unlinkat mog_unlinkpath
#endif

static int mog_renamepath(struct mog_svc *svc, const char *old, const char *new)
{
	char fsnew[MY_PATHMAX];
	char fsold[MY_PATHMAX];

	GET_FSPATH(fsold, old);
	GET_FSPATH(fsnew, new);
	return rename(fsold, fsnew);
}
#ifdef HAVE_RENAMEAT
static int mog_renameat(struct mog_svc *svc, const char *old, const char *new)
{
	int rc = renameat(svc->dirfd, old + 1, svc->dirfd, new + 1);

	if (rc < 0 && errno == ENOSYS)
		return mog_renamepath(svc, old, new);
	return rc;
}
#else
#  define mog_renameat mog_renamepath
#endif

static int mog_mkdirpath(struct mog_svc *svc, const char *path, mode_t mode)
{
	char fspath[MY_PATHMAX];

	GET_FSPATH(fspath, path);
	return mkdir(fspath, mode);
}
#ifdef HAVE_MKDIRAT
static int mog_mkdirat(struct mog_svc *svc, const char *path, mode_t mode)
{
	int rc = mkdirat(svc->dirfd, path + 1, mode);

	if (rc < 0 && errno == ENOSYS)
		return mog_mkdirpath(svc, path, mode);
	return rc;
}
#else
#  define mog_mkdirat mog_mkdirpath
#endif

int (*mog_stat)(struct mog_svc *, const char *, struct stat *) = mog_statat;
int (*mog_open_read)(struct mog_svc *, const char *) = mog_openat_read;
int (*mog_open_put)(struct mog_svc *, const char *, int) = mog_openat_put;
int (*mog_unlink)(struct mog_svc *, const char *) = mog_unlinkat;
int (*mog_rename)(struct mog_svc *, const char *, const char *) = mog_renameat;
int (*mog_mkdir)(struct mog_svc *, const char *, mode_t) = mog_mkdirat;
