/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
#include "cmogstored.h"
#include "http.h"

/* stash any pipelined data for the next round */
static void
http_defer_rbuf(struct mog_http *http, struct mog_rbuf *rbuf, size_t buf_len)
{
	struct mog_rbuf *old = http->rbuf;
	size_t defer_bytes = buf_len - http->offset;
	char *src = rbuf->rptr + http->offset;

	if (http->skip_rbuf_defer) {
		http->skip_rbuf_defer = 0;
		return;
	}

	assert(http->offset >= 0 && "http->offset negative");
	assert(defer_bytes <= MOG_HTTP_RBUF_SIZE && "defer bytes overflow");

	if (defer_bytes == 0) {
		mog_rbuf_free_and_null(&http->rbuf);
	} else if (old) { /* no allocation needed, reuse existing */
		assert(old == rbuf && "http->rbuf not reused properly");
		memmove(old->rptr, src, defer_bytes);
		old->rsize = defer_bytes;
	} else {
		http->rbuf = mog_rbuf_new(MOG_HTTP_RBUF_SIZE);
		memcpy(http->rbuf->rptr, src, defer_bytes);

		/* DANGER: we'll set this back to MOG_HTTP_RBUF_SIZE later: */
		http->rbuf->rsize = defer_bytes;
	}
	http->offset = 0;
}

static void
http_process_client(struct mog_http *http, char *buf, size_t buf_len)
{
	switch (http->http_method) {
	case MOG_HTTP_METHOD_NONE: assert(0 && "BUG: unset HTTP method");
	case MOG_HTTP_METHOD_GET: mog_http_get_open(http, buf, false); break;
	case MOG_HTTP_METHOD_HEAD: mog_http_get_open(http, buf, true); break;
	case MOG_HTTP_METHOD_DELETE: mog_http_delete(http, buf); break;
	case MOG_HTTP_METHOD_MKCOL: mog_http_mkcol(http, buf); break;
	case MOG_HTTP_METHOD_PUT: mog_http_put(http, buf, buf_len); break;
	}
}

MOG_NOINLINE static void http_close(struct mog_fd *mfd)
{
	struct mog_http *http = &mfd->as.http;

	mog_rbuf_free(http->rbuf);
	assert((http->wbuf == NULL || http->wbuf == MOG_WR_ERROR) &&
	       "would leak http->wbuf on close");
	mog_close(mfd->fd);
}

/* returns true if we can continue queue step, false if not */
static enum mog_next http_wbuf_in_progress(struct mog_http *http)
{
	assert(http->wbuf != MOG_WR_ERROR && "still active after write error");
	switch (mog_tryflush(mog_fd_of(http)->fd, &http->wbuf)) {
	case MOG_WRSTATE_ERR:
		return MOG_NEXT_CLOSE;
	case MOG_WRSTATE_DONE:
		if (!http->persistent) return MOG_NEXT_CLOSE;
		if (http->forward == NULL)
			mog_http_reset_parser(http);
		assert(http->offset == 0 && "bad offset");
		return MOG_NEXT_ACTIVE;
	case MOG_WRSTATE_BUSY:
		/* unlikely, we never put anything big in wbuf */
		return MOG_NEXT_WAIT_WR;
	}
	assert(0 && "compiler bug?");
	return MOG_NEXT_CLOSE;
}

static enum mog_next http_forward_in_progress(struct mog_fd *mfd)
{
	enum mog_http_method method = mfd->as.http.http_method;

	if (method == MOG_HTTP_METHOD_GET)
		return mog_http_get_in_progress(mfd);

	assert(method == MOG_HTTP_METHOD_PUT && "bad http_method for forward");

	return mog_http_put_in_progress(mfd);
}

static enum mog_next http_queue_step(struct mog_fd *mfd)
{
	static const size_t capa = MOG_HTTP_RBUF_SIZE;
	struct mog_http *http = &mfd->as.http;
	struct mog_rbuf *rbuf;
	char *buf;
	ssize_t r;
	off_t off;
	size_t buf_len = 0;
	enum mog_parser_state state;

	assert(mfd->fd >= 0 && "http fd is invalid");

	if (http->wbuf) return http_wbuf_in_progress(http);
	if (http->forward) return http_forward_in_progress(mfd);

	/* we may have pipelined data in http->rbuf */
	rbuf = http->rbuf ? http->rbuf : mog_rbuf_get(capa);
	buf = rbuf->rptr;
	off = http->offset;
	assert(off >= 0 && "offset is negative");
	assert(off < capa && "offset is too big");
	if (http->rbuf) {
		/* request got pipelined, resuming now */
		buf_len = http->rbuf->rsize;
		http->rbuf->rsize = capa; /* DANGER see http_defer_rbuf */
		assert(http->offset <= buf_len && "bad offset from pipelining");
		if (http->offset < buf_len)
			goto parse;
	}
reread:
	r = read(mfd->fd, buf + off, capa - off);
	if (r > 0) {
		buf_len = r + off;
parse:
		state = mog_http_parse(http, buf, buf_len);

		switch (state) {
		case MOG_PARSER_ERROR:
			mog_http_resp(http, "400 Bad Request", false);
			return MOG_NEXT_CLOSE;
		case MOG_PARSER_CONTINUE:
			assert(http->wbuf == NULL &&
			       "tried to write (and failed) with partial req");
			if (http->offset == capa) {
				assert(buf_len == capa && "bad rbuf");
				syslog(LOG_ERR, "http request too large");
				return MOG_NEXT_CLOSE;
			}
			off = http->offset;
			goto reread;
		case MOG_PARSER_DONE:
			http_process_client(http, buf, buf_len);
			if (http->wbuf == MOG_WR_ERROR)
				return MOG_NEXT_CLOSE;
			if (http->wbuf) {
				http_defer_rbuf(http, rbuf, buf_len);
				return MOG_NEXT_WAIT_WR;
			} else if (http->forward) {
				http_defer_rbuf(http, rbuf, buf_len);
			} else if (!http->persistent) {
				return MOG_NEXT_CLOSE;
			} else {
				http_defer_rbuf(http, rbuf, buf_len);
				mog_http_reset_parser(http);
			}
			return MOG_NEXT_ACTIVE;
		}
	} else if (r == 0) { /* client shut down */
		return MOG_NEXT_CLOSE;
	} else {
		switch (errno) {
		case_EAGAIN:
			if (buf_len > 0) {
				if (http->rbuf == NULL)
					http->rbuf = mog_rbuf_defer(rbuf);
				http->rbuf->rsize = buf_len; /* DANGER */
			}
			return MOG_NEXT_WAIT_RD;
		case EINTR: goto reread;
		case ECONNRESET:
		case ENOTCONN:
			return MOG_NEXT_CLOSE;
		default:
			syslog(LOG_NOTICE, "http client died: %m");
			return MOG_NEXT_CLOSE;
		}
	}

	assert(0 && "compiler bug?");
	return MOG_NEXT_CLOSE;
}

void mog_http_queue_step(struct mog_fd *mfd)
{
	struct mog_queue *q = mfd->as.http.svc->queue;

	/* centralize all queue transitions here: */
	switch (http_queue_step(mfd)) {
	case MOG_NEXT_CLOSE: http_close(mfd); return;
	case MOG_NEXT_ACTIVE: mog_activeq_push(q, mfd); return;
	case MOG_NEXT_WAIT_RD: mog_idleq_push(q, mfd, MOG_QEV_RD); return;
	case MOG_NEXT_WAIT_WR: mog_idleq_push(q, mfd, MOG_QEV_WR); return;
	case MOG_NEXT_IGNORE:
		assert(0 && "refused to put HTTP client into ignore state");
	}
}

/* called immediately after accept(), this initializes the mfd (once) */
void mog_http_post_accept(struct mog_fd *mfd, struct mog_svc *svc)
{
	struct mog_http *http = &mfd->as.http;

	mfd->fd_type = MOG_FD_TYPE_HTTP;
	assert((mfd->in_queue = 0) == 0 && "in_queue check");
	mog_http_init(http, svc);
	mog_http_queue_step(mfd);
}

/*
 * returns a NUL-terminated HTTP path from the rbuf pointer
 * returns NUL if we got a bad path.
 */
char *mog_http_path(struct mog_http *http, char *buf)
{
	char *path = buf + http->path_tip;
	size_t len = http->path_end - http->path_tip;

	assert(http->path_end > http->path_tip && "bad HTTP path from parser");

	if (! mog_valid_path(path, len))
		return NULL;

	path[len] = '\0';

	return path;
}


/* TODO: see if the iovec overheads of writev() is even worth it... */
void
mog_http_resp0(
	struct mog_http *http,
	struct iovec *status,
	bool alive)
{
	struct iovec iov;
	struct mog_now *now;
	char *dst = iov.iov_base = mog_fsbuf_get(&iov.iov_len);

	assert(status->iov_len * 2 + 1024 < iov.iov_len && "fsbuf too small");

#define CPY(str) mempcpy(dst, (str),(sizeof(str)-1))
	dst = CPY("HTTP/1.1 ");
	dst = mempcpy(dst, status->iov_base, status->iov_len);
	dst = CPY("\r\nStatus: ");
	dst = mempcpy(dst, status->iov_base, status->iov_len);
	dst = CPY("\r\nDate: ");
	now = mog_now();
	dst = mempcpy(dst, now->httpdate, sizeof(now->httpdate)-1);
	if (alive && http->persistent) {
		dst = CPY("\r\nContent-Length: 0"
			"\r\nContent-Type: text/plain"
			"\r\nConnection: keep-alive\r\n\r\n");
	} else {
		http->persistent = 0;
		dst = CPY("\r\nContent-Length: 0"
			"\r\nContent-Type: text/plain"
			"\r\nConnection: close\r\n\r\n");
	}
	iov.iov_len = dst - (char *)iov.iov_base;
	assert(http->wbuf == NULL && "tried to write with wbuf");

	http->wbuf = mog_trywritev(mog_fd_of(http)->fd, &iov, 1);
}
