/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
/* kqueue-specific parts see queue_common.h and activeq.c for the rest */
#include "queue_common.h"
/*
 * a poll/select/libev/libevent-based implementation would have a hard time
 * migrating clients between threads
 */
#ifdef HAVE_KQUEUE
static const uintptr_t MOG_XFER_IDENT = (uintptr_t)INT_MAX + 1;

struct mog_queue * mog_queue_new(void)
{
	int kqueue_fd = kqueue();
	struct kevent event;
	int rc;

	if (kqueue_fd < 0)
		die("kqueue failed: %s\n", strerror(errno));

	EV_SET(&event, MOG_XFER_IDENT, EVFILT_USER,
               EV_ADD | EV_CLEAR, 0, 0, NULL);
	do {
		rc = kevent(kqueue_fd, &event, 1, NULL, 0, NULL);
	} while (rc < 0 && errno == EINTR);

	if (rc != 0)
		die("failed to add MOG_XFER_IDENT event: %s", strerror(errno));

	return mog_queue_init(kqueue_fd);
}

/*
 * grabs one active event off the event queue
 */
struct mog_fd * mog_idleq_wait(struct mog_queue *q, int dontwait)
{
	int rc;
	struct mog_fd *mfd;
	struct kevent event;
	static const struct timespec ts = { 0, 0 };

retry:
	/* TODO: is kqueue a cancellation point? */
	pthread_testcancel();
	rc = kevent(q->queue_fd, NULL, 0, &event, 1, dontwait ? &ts : NULL);
	switch (rc) {
	case 1:
		if (event.ident == MOG_XFER_IDENT) {
			mfd = mog_activeq_trytake(q);
			if (mfd || dontwait)
				return mfd;
			goto retry;
		}
		mfd = event.udata;
		mog_fd_check_out(mfd);
		return mfd;
	case 0:
		if (dontwait)
			return NULL;
		goto retry;
	}

	switch (errno) {
	case EINTR:
		goto retry;
	case EBADF:
	case EINVAL: /* kqueue_fd can be hit */
		return NULL;
	}

	die("kevent(wait) failed with (%d): %s\n", rc, strerror(errno));
	return NULL;
}

MOG_NOINLINE static void
kqueue_add_error(struct mog_queue *q, struct mog_fd *mfd)
{
	switch (errno) {
	case EBADF:
		/* TODO: check for shutdown races */
		syslog(LOG_ERR, "bad file descriptor for kevent(EV_ADD)");
		return;
	case ENOMEM:
		syslog(LOG_ERR, "kevent(EV_ADD) out-of-space");
		if (mfd) {
			syslog(LOG_ERR, "falling back to active queue");
			mog_activeq_push(q, mfd);
		}
		return;
	default:
		syslog(LOG_ERR, "unhandled kqueue(EV_ADD) error: %m");
		assert(0 && "BUG in our usage of kqueue");
	}
}

static void qxfer(struct mog_queue *q, struct mog_fd *mfd)
{
	struct kevent event;
	int rc;

	assert(mfd->queue_state == MOG_QUEUE_STATE_NEW && "bad queue state");
	mfd->queue_state = MOG_QUEUE_STATE_OLD;

	mog_fd_check_in(mfd);
	CHECK(int, 0, pthread_mutex_lock(&q->activeq_lock));
	SIMPLEQ_INSERT_TAIL(&q->activeq_head, mfd, active_fd);
	CHECK(int, 0, pthread_mutex_unlock(&q->activeq_lock));

	EV_SET(&event, MOG_XFER_IDENT, EVFILT_USER, 0, NOTE_TRIGGER, 0, NULL);
	do {
		rc = kevent(q->queue_fd, &event, 1, NULL, 0, NULL);
	} while (rc < 0 && errno == EINTR);

	if (rc != 0)
		kqueue_add_error(q, NULL);
}

static void qpush(struct mog_queue *q, struct mog_fd *mfd, enum mog_qev ev)
{
	struct kevent event;
	int rc;
	u_short flags = EV_ADD | EV_ONESHOT;

	mfd->queue_state = MOG_QUEUE_STATE_OLD;
	EV_SET(&event, mfd->fd, (short)ev, flags, 0, 0, mfd);

	mog_fd_check_in(mfd);
	do {
		rc = kevent(q->queue_fd, &event, 1, NULL, 0, NULL);
	} while (rc < 0 && errno == EINTR);

	if (rc != 0) {
		mog_fd_check_out(mfd);
		kqueue_add_error(q, mfd);
	}
}

/*
 * Pushes in one mog_fd for kqueue to watch.
 *
 * Only call this from the mog_accept_loop *or*
 * if EAGAIN/EWOULDBLOCK is encountered in mog_queue_loop.
 */
void mog_idleq_push(struct mog_queue *q, struct mog_fd *mfd, enum mog_qev ev)
{
	ev == MOG_QEV_RW ? qxfer(q, mfd) : qpush(q, mfd, ev);
}
#else /* ! HAVE_KQUEUE */
typedef int avoid_empty_file;
#endif /* ! HAVE_KQUEUE */
